/**
 * @file application.cpp
 * @brief Application class
 *
 * @author Eric Butler (edbutler)
 */

#include "application/application.hpp"

#include "opengl.hpp"
#include "application/imageio.hpp"

#include <GL/freeglut.h>

#include <cassert>
#include <cstdlib>
#include <iostream>
#include <map>

namespace _462 {


// types and static data



struct Dimension
{
    int width;
    int height;
};

struct MenuState
{
    typedef std::map< int, Window::MenuCallback > MenuItemMap;

    // glut id of this menu
    int id;
    MenuItemMap items;
    int last_item_id;

    MenuState() : id( 0 ), last_item_id( 0 ) { }
};

struct WindowState
{
    // screen dimension in pixels
    Dimension screen;

    MenuState menu;

    // number of seconds in a single frame
    real_t period;

    // the current application
    Application* app;
    // window handle to pass to objects
    Window* window;

    WindowState()
        : period( 0 ), app( 0 ), window( 0 ) { }
};

static WindowState* state = 0;


static void display_callback()
{
    // render the application
    assert( state && state->app );
    state->app->render( state->window );
    // flush and swap the buffer
    glFlush();
    glutSwapBuffers();
}

static void window_resize_callback( int width, int height )
{
    // update window information
    state->screen.width = width;
    state->screen.height = height;
}

static void timer_callback( int millis )
{
    assert( state && state->app );

    // set the next timer callback to occur after the period has elapsed
    glutTimerFunc( millis, &timer_callback, millis );
    // update the application
    state->app->update( state->window, state->period );
    // set the display to update
    glutPostRedisplay();
}

static void menu_callback( int menu_item )
{
    assert( state );
    MenuState::MenuItemMap::const_iterator iter = state->menu.items.find( menu_item );
    assert( iter != state->menu.items.end() );
    iter->second( state->app, state->window );
}

static size_t get_modifiers()
{
    int glut_modifiers = glutGetModifiers();
    return
        ( GLUT_ACTIVE_SHIFT & glut_modifiers ? Application::MODIFIER_SHIFT : 0 ) |
        ( GLUT_ACTIVE_CTRL & glut_modifiers ? Application::MODIFIER_CTRL : 0 ) |
        ( GLUT_ACTIVE_ALT & glut_modifiers ? Application::MODIFIER_ALT : 0 );
}

static void mouse_button_callback( int glut_button, int pressed, int x, int y )
{
    assert( state && state->app );

    Application::MouseButton app_button;
    switch ( glut_button ) {
        case GLUT_LEFT_BUTTON:
            app_button = Application::MOUSE_LEFT;
            break;
        case GLUT_MIDDLE_BUTTON:
            app_button = Application::MOUSE_MIDDLE;
            break;
        case GLUT_RIGHT_BUTTON:
            app_button = Application::MOUSE_RIGHT;
            break;
        default:
            break;
    }

    state->app->on_mouse_press( state->window, app_button, x, y, get_modifiers(), pressed == GLUT_DOWN );
}

static void mouse_motion_callback( int x, int y )
{
    assert( state && state->app );
    state->app->on_mouse_move( state->window, x, y );
}

static void keyboard_callback( unsigned char key, int x, int y )
{
    assert( state && state->app );
    bool pressed = true; // there are no up events with glut
    state->app->on_key_press( state->window, char( key ), get_modifiers(), pressed );
}

static void terminate_application( Application*, Window* )
{
    glutLeaveMainLoop();
}


// header function definitions


Window::Window() { }

Window::~Window() { }

static bool initialize_window( int width, int height, const char* title )
{
    assert( state );

    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
    glutInitWindowSize( width, height );
    if ( glutCreateWindow( title ) == 0 )
        return false;

    // callback for rendering
    glutDisplayFunc( display_callback );
    // callback for window size changes
    glutReshapeFunc( window_resize_callback );
    // will initialize values
    window_resize_callback( width, height );

    return true;
}

static bool initialize_gui()
{
    state->menu.id = glutCreateMenu( menu_callback );
    glutSetMenu( state->menu.id );

    // callback for mouse drags
    glutMotionFunc( mouse_motion_callback );
    // callback for mouse drags with no button pressed
    glutPassiveMotionFunc( mouse_motion_callback );
    // callback for mouse button events
    glutMouseFunc( mouse_button_callback );
    // callback for keyboard
    glutKeyboardFunc( keyboard_callback );

    glutAttachMenu( GLUT_RIGHT_BUTTON );

    return true;
}

void Window::start_application( Application* app,
                                int argc, char *argv[],
                                int width, int height, real_t fps,
                                const char* title )
{
    assert( app );
    // only create an app the first time this function is called
    if ( state )
        return;


    int period = static_cast< int >( 1000.0 / fps ); // frame period in milliseconds


    // initialize the application state
    state = new WindowState();
    if ( !state )
        goto FAIL;
    state->app = app;
    state->window = new Window();
    if ( !state->window )
        goto FAIL;
    state->period = 1.0 / fps;


    // init glut and parse out other args
    glutInit( &argc, argv );
    glutSetOption( GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_CONTINUE_EXECUTION );


    if ( !initialize_window( width, height, title ) )
        goto FAIL;

    if ( !initialize_gui() )
        goto FAIL;


    // initialize the application
    if ( !app->initialize( state->window, argc, argv ) )
        goto FAIL;


    // add default menu items
    state->window->add_menu_item( "Save Screenshot", screenshot_callback );
    state->window->add_menu_item( "Quit", _462::terminate_application );


    // start the "update loop" by hacking the timer callback
    glutTimerFunc( period, &timer_callback, period );


    // start main loop
    glutMainLoop();


    // clean up and exit
    assert( state->app );
    state->app->destroy( state->window );
    return;


  FAIL:
    std::cerr << "Failed to start applcation, aborting\n";
    exit( 1 );
}

void Window::get_dimension( int* width, int* height )
{
    assert( width && height );
    assert( state );
    *width = state->screen.width;
    *height = state->screen.height;
}

void Window::add_menu_item( const char* name, MenuCallback callback )
{
    int item_id = state->menu.last_item_id++;
    // do this first in case insert throws exception
    state->menu.items.insert( std::make_pair( item_id, callback ) );
    glutAddMenuEntry( name, item_id );
}

void Window::terminate_application()
{
    assert( state );
    _462::terminate_application( state->app, state->window );
}

} /* _462 */

