/**
 * @file main0.cpp
 * @brief The main function and application for p0.
 *
 * @author Eric Butler (edbutler)
 */

/*
    EDIT THIS FILE FOR P0.
 */

#include "application/application.hpp"

#include "opengl.hpp"
#include "p0.hpp"
#include "math/vector.hpp"
#include "scene/camera.hpp"
#include "scene/model.hpp"
#include <iostream>

namespace _462 {

#define CAMERA_TRANSLATION_SCALE_FACTOR 0.01
#define CAMERA_ROTATION_SCALE_FACTOR 0.01

/* different camera operations performed by gui */
enum CameraAction
{
    CA_NONE, // represents event of doing nothing
    CA_FOCUS_PITCH,
    CA_FOCUS_YAW,
    CA_FOCUS_ZOOM,
    CA_TRANSLATE_X,
    CA_TRANSLATE_Y,
    CA_TRANSLATE_Z,
    CA_ROLL,
    CA_PITCH,
    CA_YAW
};

struct MouseState
{
    // mouse position as of last callback
    int x,y;
    // button state of the left, middle, and right buttons. true means
    // down, false means up
    bool button_state[3];

    /* The current camera control state of the x and y screen axes. */
    CameraAction camera_control[2];

    MouseState():
        x(0), y(0)
    {
        button_state[Application::MOUSE_LEFT] = false;
        button_state[Application::MOUSE_MIDDLE] = false;
        button_state[Application::MOUSE_RIGHT] = false;
        camera_control[0] = CA_NONE;
        camera_control[1] = CA_NONE;
    }
};

enum {
    MODIFIER_INDEX_NONE = 0,
    MODIFIER_INDEX_CTRL,
    MODIFIER_INDEX_SHIFT
};

/*
    Camera controls:
    left-click, x-axis yaws around focus
    left-click, y-axis pitches around focus
    middle-click, y-axis zooms toward/away from focus

    ctrl + left-click, x-axis controls translation along local x-axis
    ctrl + left-click, y-axis controls translation along local y-axis
    ctrl + middle-click, y-axis controls translation along local z-axis

    shift + left-click, x-axis controls yaw
    shift + left-click, y-axis controls pitch
    shift + middle-click, y-axis controls roll

    array elements are indexed by
    control_map[modifier key state][button][mouse axis]
 */
static const CameraAction control_map[3][2][2] = {
    {
        { CA_FOCUS_YAW  , CA_FOCUS_PITCH },
        { CA_NONE       , CA_FOCUS_ZOOM  }
    },
    {
        { CA_TRANSLATE_X, CA_TRANSLATE_Y },
        { CA_NONE       , CA_TRANSLATE_Z }
    },
    {
        { CA_YAW        , CA_PITCH       },
        { CA_NONE       , CA_ROLL,       }
    }
};


// callback functions and helper functions


class P0Application : public Application
{
public:

    P0Application() : initialized( false ) { }
    virtual ~P0Application() { }

	virtual bool initialize( Window* window, int argc, char* argv[] );
	virtual void destroy( Window* window );
	virtual void update( Window* window, real_t time );
	virtual void render( Window* window );

    virtual void on_mouse_press( Window* window, MouseButton button, int x, int y, size_t modifiers, bool pressed );
    virtual void on_mouse_move( Window* window, int x, int y );

private:

    void apply_control( CameraAction ca, int delta );

    MouseState mouse;

    Camera camera;

    Model model;

    bool initialized;
};

template< PolygonMode MODE >
static void polygon_mode_callback( Application*,  Window* )
{
    prj_set_polygon_mode( MODE );
}

static void print_usage( const char* prog_name )
{
    std::cout << "Usage: "
        << prog_name
        << " [FILE]\n"
        << "Loads a model from FILE and displays it."
        << std::endl;
}

bool P0Application::initialize( Window* window, int argc, char* argv[] )
{
    // make sure not already initialized
    assert( !initialized );
    initialized = true;

    // get the file from the first arg
    assert( argc > 0 );
    if ( argc != 2 ) {
        print_usage( argv[0] );
        return false;
    }

    if ( !model.load_from_obj( argv[1] ) ) {
        std::cerr << "Error loading model. Aborting." << std::endl;
        return false;
    }

    // change the pixel alignment requirements
    glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
    glPixelStorei( GL_PACK_ALIGNMENT, 1 );

    // initialize the project
    _462::prj_initialize( &camera, &model );

    // set the right-click callbacks
    window->add_menu_item( "Draw Points", &polygon_mode_callback< POLYMODE_POINT > );
    window->add_menu_item( "Draw Lines", &polygon_mode_callback< POLYMODE_LINE > );
    window->add_menu_item( "Draw Fill", &polygon_mode_callback< POLYMODE_FILL > );

    return true;
}

void P0Application::destroy( Window* )
{
    // make sure not called before initializing
    assert( initialized );
    _462::prj_destroy( &model );
}

void P0Application::update( Window*, real_t )
{
    // nothing to update
}

void P0Application::render( Window* window )
{
    int width, height;

    // query current window size, resize viewport
    window->get_dimension( &width, &height );
    glViewport( 0, 0, width, height );

    // fix camera aspect
    camera.aspect = real_t( width ) / real_t( height );

    _462::prj_render( &camera, &model );
}

void P0Application::on_mouse_press( Window* window, MouseButton button, int x, int y, size_t modifiers, bool pressed )
{
    mouse.button_state[button] = pressed;

    // left and middle button control camera
    if ( button != MOUSE_RIGHT ) {
        // if either is released, cancel controls
        if ( !pressed ) {
            mouse.camera_control[0] = CA_NONE;
            mouse.camera_control[1] = CA_NONE;
        } else {
            int modifier_index;

            if ( modifiers & MODIFIER_CTRL )
                modifier_index = MODIFIER_INDEX_CTRL;
            else if ( modifiers & MODIFIER_SHIFT )
                modifier_index = MODIFIER_INDEX_SHIFT;
            else
                modifier_index = MODIFIER_INDEX_NONE;

            mouse.camera_control[0] = control_map[modifier_index][button][0];
            mouse.camera_control[1] = control_map[modifier_index][button][1];
        }
    }
}

void P0Application::on_mouse_move( Window* window, int x, int y )
{
    int oldx = mouse.x;
    int oldy = mouse.y;

    mouse.x = x;
    mouse.y = y;

    if ( x != oldx )
        apply_control( mouse.camera_control[0], x - oldx );
    if ( y != oldy )
        apply_control( mouse.camera_control[1], y - oldy );
}

/**
 * Applies the given control to the current scene's camera.
 * @param ca The control to apply.
 * @param delta The distance of mouse movement.
 */
void P0Application::apply_control( CameraAction ca, int delta )
{
    // if nothing, just return
    if ( ca == CA_NONE )
        return;

    real_t angle = CAMERA_ROTATION_SCALE_FACTOR * delta;
    real_t distance = CAMERA_TRANSLATION_SCALE_FACTOR * delta;
    real_t new_focus_dist;

    switch ( ca )
    {
    case CA_FOCUS_YAW:
        camera.yaw_about_focus( -angle );
        break;

    case CA_FOCUS_PITCH:
        camera.pitch_about_focus( -angle );
        break;

    case CA_FOCUS_ZOOM:
        // change focus based on an exponential scale
        new_focus_dist = pow( 2, distance ) * camera.focus_dist;
        camera.translate( Vector3::UnitZ * ( new_focus_dist - camera.focus_dist ) );
        camera.focus_dist = new_focus_dist;
        break;

    case CA_TRANSLATE_X:
        camera.translate( -Vector3::UnitX * distance );
        break;

    case CA_TRANSLATE_Y:
        camera.translate( Vector3::UnitY * distance );
        break;

    case CA_TRANSLATE_Z:
        camera.translate( Vector3::UnitZ * distance );
        break;

    case CA_ROLL:
        camera.roll( angle );
        break;

    case CA_PITCH:
        camera.pitch( angle );
        break;

    case CA_YAW:
        camera.yaw( angle );
        break;

    default:
        break;
    }
}

} /* _462 */

using namespace _462;

/**
 * Application entry point.
 */
int main( int argc, char* argv[] )
{
    int width = 800;
    int height = 600;
    real_t fps = 30.0;
    const char* title = "462 Project 0";

    // start a new application
    Window::start_application( new _462::P0Application(),
                               argc, argv, width, height, fps, title );
}

