/**
 * @file matrix.cpp
 * @brief Matrix classes.
 *
 * @author Eric Butler (edbutler)
 * @author Zeyang Li (zeyangl)
 */

#include "math/matrix.hpp"

#include "math/quaternion.hpp"
#include <cstring>

namespace _462 {

const Matrix3 Matrix3::Identity = Matrix3( 1, 0, 0,
                                           0, 1, 0,
                                           0, 0, 1 );

const Matrix3 Matrix3::Zero = Matrix3( 0, 0, 0,
                                       0, 0, 0,
                                       0, 0, 0 );


Matrix3::Matrix3( real_t r[SIZE] )
{
    memcpy( m, r, sizeof r );
}

Matrix3::Matrix3( real_t m00, real_t m10, real_t m20,
                  real_t m01, real_t m11, real_t m21,
                  real_t m02, real_t m12, real_t m22 )
{
    _m[0][0] = m00;
    _m[1][0] = m10;
    _m[2][0] = m20;
    _m[0][1] = m01;
    _m[1][1] = m11;
    _m[2][1] = m21;
    _m[0][2] = m02;
    _m[1][2] = m12;
    _m[2][2] = m22;
}

Matrix3 Matrix3::from_columns( const Vector3& x,
                               const Vector3& y,
                               const Vector3& z )
{
    return Matrix3(
        x.x, y.x, z.x,
        x.y, y.y, z.y,
        x.z, y.z, z.z
    );
}

Matrix3 Matrix3::operator+( const Matrix3& rhs ) const
{
    Matrix3 rv;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = m[i] + rhs.m[i];
    return rv;
}

Matrix3& Matrix3::operator+=( const Matrix3& rhs )
{
    for ( int i = 0; i < SIZE; i++ )
        m[i] += rhs.m[i];
    return *this;
}

Matrix3 Matrix3::operator-( const Matrix3& rhs ) const
{
    Matrix3 rv;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = m[i] - rhs.m[i];
    return rv;
}

Matrix3& Matrix3::operator-=( const Matrix3& rhs )
{
    for ( int i = 0; i < SIZE; i++ )
        m[i] -= rhs.m[i];
    return *this;
}

Matrix3 Matrix3::operator*( const Matrix3& rhs ) const
{
    Matrix3 product;
    for ( int i = 0; i < DIM; ++i )
        for ( int j = 0; j < DIM; ++j )
            product._m[i][j] =
                _m[0][j] * rhs._m[i][0] + _m[1][j] * rhs._m[i][1] +
                _m[2][j] * rhs._m[i][2];
    return product;
}

Vector3 Matrix3::operator*( const Vector3& v ) const
{
    return Vector3( _m[0][0]*v.x + _m[1][0]*v.y + _m[2][0]*v.z,
                    _m[0][1]*v.x + _m[1][1]*v.y + _m[2][1]*v.z,
                    _m[0][2]*v.x + _m[1][2]*v.y + _m[2][2]*v.z );
}

Matrix3& Matrix3::operator*=( const Matrix3& rhs )
{
    return *this = operator*( rhs );
}

Matrix3 Matrix3::operator*( real_t r ) const
{
    Matrix3 rv;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = m[i] * r;
    return rv;
}

Matrix3& Matrix3::operator*=( real_t r )
{
    for ( int i = 0; i < SIZE; i++ )
        m[i] *= r;
    return *this;
}

Matrix3 Matrix3::operator/( real_t r ) const
{
    Matrix3 rv;
    real_t inv = 1 / r;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = m[i] * inv;
    return rv;
}

Matrix3& Matrix3::operator/=( real_t r )
{
    real_t inv = 1 / r;
    for ( int i = 0; i < SIZE; i++ )
        m[i] *= inv;
    return *this;
}

Matrix3 Matrix3::operator-() const
{
    Matrix3 rv;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = -m[i];
    return rv;
}

bool Matrix3::operator==( const Matrix3& rhs ) const
{
    return memcmp( m, rhs.m, sizeof m ) == 0;
}

bool Matrix3::operator!=( const Matrix3& rhs ) const
{
    return !operator==( rhs );
}

void Matrix3::transpose( Matrix3& rv ) const
{
    rv = Matrix3( _m[0][0], _m[0][1], _m[0][2],
                  _m[1][0], _m[1][1], _m[1][2],
                  _m[2][0], _m[2][1], _m[2][2] );
}

void Matrix3::inverse( Matrix3& rv ) const
{
    rv._m[0][0] = _m[1][1] * _m[2][2] - _m[1][2] * _m[2][1];
    rv._m[0][1] = _m[0][2] * _m[2][1] - _m[0][1] * _m[2][2];
    rv._m[0][2] = _m[0][1] * _m[1][2] - _m[0][2] * _m[1][1];
    rv._m[1][0] = _m[1][2] * _m[2][0] - _m[1][0] * _m[2][2];
    rv._m[1][1] = _m[0][0] * _m[2][2] - _m[0][2] * _m[2][0];
    rv._m[1][2] = _m[0][2] * _m[1][0] - _m[0][0] * _m[1][2];
    rv._m[2][0] = _m[1][0] * _m[2][1] - _m[1][1] * _m[2][0];
    rv._m[2][1] = _m[0][1] * _m[2][0] - _m[0][0] * _m[2][1];
    rv._m[2][2] = _m[0][0] * _m[1][1] - _m[0][1] * _m[1][0];

    real_t det = _m[0][0] * rv._m[0][0] +
                 _m[0][1] * rv._m[1][0] +
                 _m[0][2] * rv._m[2][0];

    real_t invdet = 1.0 / det;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] *= invdet;
}

const Matrix4 Matrix4::Identity = Matrix4( 1, 0, 0, 0,
                                  0, 1, 0, 0,
                                  0, 0, 1, 0,
                                  0, 0 , 0, 1 );

const Matrix4 Matrix4::Zero = Matrix4( 0, 0, 0, 0,
                                       0, 0, 0, 0,
                                       0, 0, 0, 0,
                                       0, 0, 0, 0 );

Matrix4::Matrix4( real_t r[SIZE] )
{
    memcpy( m , r, sizeof r );
}

Matrix4::Matrix4( real_t m00, real_t m10, real_t m20, real_t m30,
                  real_t m01, real_t m11, real_t m21, real_t m31,
                  real_t m02, real_t m12, real_t m22, real_t m32,
                  real_t m03, real_t m13, real_t m23, real_t m33 )
{
    _m[0][0] = m00;
    _m[1][0] = m10;
    _m[2][0] = m20;
    _m[3][0] = m30;
    _m[0][1] = m01;
    _m[1][1] = m11;
    _m[2][1] = m21;
    _m[3][1] = m31;
    _m[0][2] = m02;
    _m[1][2] = m12;
    _m[2][2] = m22;
    _m[3][2] = m32;
    _m[0][3] = m03;
    _m[1][3] = m13;
    _m[2][3] = m23;
    _m[3][3] = m33;
}


Matrix4 Matrix4::operator+( const Matrix4& rhs ) const
{
    Matrix4 rv;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = m[i] + rhs.m[i];
    return rv;
}

Matrix4& Matrix4::operator+=( const Matrix4& rhs )
{
    for ( int i = 0; i < SIZE; i++ )
        m[i] += rhs.m[i];
    return *this;
}

Matrix4 Matrix4::operator-( const Matrix4& rhs ) const
{
    Matrix4 rv;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = m[i] - rhs.m[i];
    return rv;
}

Matrix4& Matrix4::operator-=( const Matrix4& rhs )
{
    for ( int i = 0; i < SIZE; i++ )
        m[i] -= rhs.m[i];
    return *this;
}

Matrix4 Matrix4::operator*( const Matrix4& rhs ) const
{
    Matrix4 product;
    for ( int i = 0; i < DIM; ++i )
        for ( int j = 0; j < DIM; ++j )
            product._m[i][j] =
                _m[0][j] * rhs._m[i][0] + _m[1][j] * rhs._m[i][1] +
                _m[2][j] * rhs._m[i][2] + _m[3][j] * rhs._m[i][3];
    return product;
}

Vector4 Matrix4::operator*( const Vector4& v ) const
{
    return Vector4( _m[0][0]*v.x + _m[1][0]*v.y + _m[2][0]*v.z + _m[3][0]*v.w,
                    _m[0][1]*v.x + _m[1][1]*v.y + _m[2][1]*v.z + _m[3][1]*v.w,
                    _m[0][2]*v.x + _m[1][2]*v.y + _m[2][2]*v.z + _m[3][2]*v.w,
                    _m[0][3]*v.x + _m[1][3]*v.y + _m[2][3]*v.z + _m[3][3]*v.w );
}

Matrix4& Matrix4::operator*=( const Matrix4& rhs )
{
    return *this = operator*( rhs );
}

Matrix4 Matrix4::operator*( real_t r ) const
{
    Matrix4 rv;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = m[i] * r;
    return rv;
}

Matrix4& Matrix4::operator*=( real_t r )
{
    for ( int i = 0; i < SIZE; i++ )
        m[i] *= r;
    return *this;
}

Matrix4 Matrix4::operator/( real_t r ) const
{
    Matrix4 rv;
    real_t inv = 1 / r;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = m[i] * inv;
    return rv;
}

Matrix4& Matrix4::operator/=( real_t r )
{
    real_t inv = 1 / r;
    for ( int i = 0; i < SIZE; i++ )
        m[i] *= inv;
    return *this;
}

Matrix4 Matrix4::operator-() const
{
    Matrix4 rv;
    for ( int i = 0; i < SIZE; i++ )
        rv.m[i] = -m[i];
    return rv;
}

bool Matrix4::operator==( const Matrix4& rhs ) const
{
    return memcmp( m, rhs.m, sizeof m ) == 0;
}

bool Matrix4::operator!=( const Matrix4& rhs ) const
{
    return !operator==( rhs );
}

void Matrix4::transpose( Matrix4& rv ) const
{
    rv = Matrix4( _m[0][0], _m[0][1], _m[0][2], _m[0][3],
                  _m[1][0], _m[1][1], _m[1][2], _m[1][3],
                  _m[2][0], _m[2][1], _m[2][2], _m[2][3],
                  _m[3][0], _m[3][1], _m[3][2], _m[3][3] );
}

void Matrix4::inverse( Matrix4& rv ) const
{
    real_t m00 = _m[0][0], m01 = _m[0][1], m02 = _m[0][2], m03 = _m[0][3];
    real_t m10 = _m[1][0], m11 = _m[1][1], m12 = _m[1][2], m13 = _m[1][3];
    real_t m20 = _m[2][0], m21 = _m[2][1], m22 = _m[2][2], m23 = _m[2][3];
    real_t m30 = _m[3][0], m31 = _m[3][1], m32 = _m[3][2], m33 = _m[3][3];

    real_t v0 = m20 * m31 - m21 * m30;
    real_t v1 = m20 * m32 - m22 * m30;
    real_t v2 = m20 * m33 - m23 * m30;
    real_t v3 = m21 * m32 - m22 * m31;
    real_t v4 = m21 * m33 - m23 * m31;
    real_t v5 = m22 * m33 - m23 * m32;

    real_t t00 = + ( v5 * m11 - v4 * m12 + v3 * m13 );
    real_t t10 = - ( v5 * m10 - v2 * m12 + v1 * m13 );
    real_t t20 = + ( v4 * m10 - v2 * m11 + v0 * m13 );
    real_t t30 = - ( v3 * m10 - v1 * m11 + v0 * m12 );

    real_t invDet = 1 / ( t00 * m00 + t10 * m01 + t20 * m02 + t30 * m03 );

    rv._m[0][0] = t00 * invDet;
    rv._m[0][1] = t10 * invDet;
    rv._m[0][2] = t20 * invDet;
    rv._m[0][3] = t30 * invDet;

    rv._m[1][0] = - ( v5 * m01 - v4 * m02 + v3 * m03 ) * invDet;
    rv._m[1][1] = + ( v5 * m00 - v2 * m02 + v1 * m03 ) * invDet;
    rv._m[1][2] = - ( v4 * m00 - v2 * m01 + v0 * m03 ) * invDet;
    rv._m[1][3] = + ( v3 * m00 - v1 * m01 + v0 * m02 ) * invDet;

    v0 = m10 * m31 - m11 * m30;
    v1 = m10 * m32 - m12 * m30;
    v2 = m10 * m33 - m13 * m30;
    v3 = m11 * m32 - m12 * m31;
    v4 = m11 * m33 - m13 * m31;
    v5 = m12 * m33 - m13 * m32;

    rv._m[2][0] = + ( v5 * m01 - v4 * m02 + v3 * m03 ) * invDet;
    rv._m[2][1] = - ( v5 * m00 - v2 * m02 + v1 * m03 ) * invDet;
    rv._m[2][2] = + ( v4 * m00 - v2 * m01 + v0 * m03 ) * invDet;
    rv._m[2][3] = - ( v3 * m00 - v1 * m01 + v0 * m02 ) * invDet;

    v0 = m21 * m10 - m20 * m11;
    v1 = m22 * m10 - m20 * m12;
    v2 = m23 * m10 - m20 * m13;
    v3 = m22 * m11 - m21 * m12;
    v4 = m23 * m11 - m21 * m13;
    v5 = m23 * m12 - m22 * m13;

    rv._m[3][0] = - ( v5 * m01 - v4 * m02 + v3 * m03 ) * invDet;
    rv._m[3][1] = + ( v5 * m00 - v2 * m02 + v1 * m03 ) * invDet;
    rv._m[3][2] = - ( v4 * m00 - v2 * m01 + v0 * m03 ) * invDet;
    rv._m[3][3] = + ( v3 * m00 - v1 * m01 + v0 * m02 ) * invDet;
}

std::ostream& operator << ( std::ostream& o, const Matrix4& m )
{
    o << "Matrix4(";
    for ( size_t i = 0; i < 4; ++i ) {
        o << " row" << ( unsigned )i << "{ ";
        for ( size_t j = 0; j < 4; ++j ) {
            o << m._m[i][j] << " ";
        }
        o << "}";
    }
    o << ")";
    return o;
}

void make_translation_matrix( Matrix4& mat, const Vector3& pos )
{
    mat = Matrix4(
              1.0, 0.0, 0.0, pos.x,
              0.0, 1.0, 0.0, pos.y,
              0.0, 0.0, 1.0, pos.z,
              0.0, 0.0, 0.0, 1.0 );
}

void make_rotation_matrix( Matrix4& mat, const Quaternion& ori )
{
    ori.to_matrix( mat );
}

void make_scaling_matrix( Matrix4& mat, const Vector3& scl )
{
    mat = Matrix4(
              scl.x, 0.0,   0.0,   0.0,
              0.0,   scl.y, 0.0,   0.0,
              0.0,   0.0,   scl.z, 0.0,
              0.0,   0.0,   0.0,   1.0 );
}

void make_transformation_matrix( Matrix4& mat,
                                 const Vector3& pos,
                                 const Quaternion& ori,
                                 const Vector3& scl )
{
    ori.to_matrix( mat );
    Matrix4 m;
    make_scaling_matrix( m, scl );
    mat *= m;
    mat.m[12] = pos.x;
    mat.m[13] = pos.y;
    mat.m[14] = pos.z;
}

/**
 * Returns the cofactor of the given col and row. Assumes that the matrix is
 * actually a 3d matrix.
 */
template<typename T>
static real_t _cofactor2( const T& mat, size_t col, size_t row )
{
    size_t r1 = ( row + 1 ) % 3;
    size_t r2 = ( row + 2 ) % 3;
    size_t c1 = ( col + 1 ) % 3;
    size_t c2 = ( col + 2 ) % 3;

    real_t a = mat._m[c1][r1];
    real_t b = mat._m[c2][r1];
    real_t c = mat._m[c1][r2];
    real_t d = mat._m[c2][r2];

    return b*c - a*d;
}

template<typename T>
static void _make_normal_matrix( Matrix3& rv, const T& tmat )
{
    rv._m[0][0] = -_cofactor2( tmat, 0, 0 );
    rv._m[0][1] = -_cofactor2( tmat, 1, 0 );
    rv._m[0][2] = -_cofactor2( tmat, 2, 0 );
    rv._m[1][0] = -_cofactor2( tmat, 0, 1 );
    rv._m[1][1] = -_cofactor2( tmat, 1, 1 );
    rv._m[1][2] = -_cofactor2( tmat, 2, 1 );
    rv._m[2][0] = -_cofactor2( tmat, 0, 2 );
    rv._m[2][1] = -_cofactor2( tmat, 1, 2 );
    rv._m[2][2] = -_cofactor2( tmat, 2, 2 );
}

void make_normal_matrix( Matrix3& rv, const Matrix4& tmat )
{
    _make_normal_matrix( rv, tmat );
}

void make_normal_matrix( Matrix3& rv, const Matrix3& tmat )
{
    _make_normal_matrix( rv, tmat );
}

} /* _462 */

