/**
 * @file matrix.hpp
 * @brief Matrix classes.
 *
 * @author Eric Butler (edbutler)
 * @author Zeyang Li (zeyangl)
 */

#ifndef _462_MATH_MATRIX_HPP_
#define _462_MATH_MATRIX_HPP_

#include "math/math.hpp"
#include "math/vector.hpp"
#include <cassert>

namespace _462 {

class Quaternion;

/**
 * A 3x3 matrix.
 *
 * The matrix is in column-vector compatible notation and is stored
 * in memory by column first. So the memory layout looks like
 *
 * [ m[0] m[3] m[6] ]
 * [ m[1] m[4] m[7] ]
 * [ m[2] m[5] m[8] ]
 *
 * where vectors are treated as column matrices (3x1), i.e. vectors are
 * multiplied on the right of the matrix (M * V)
 *
 * NOTE: The notation mij or m[i][j] is the ith COLUMN and the jth ROW.
 *
 * The matrix is meant for use in linear transformations. Transformations must
 * be combined right-to-left, so if you want to apply A then B the C to a
 * vector V, you would use C*B*A*V.
 */
class Matrix3
{
public:
    /**
     * The identity matrix.
     */
    static const Matrix3 Identity;

    /**
     * The zero matrix.
     */
    static const Matrix3 Zero;

    static const int DIM = 3;
    static const int SIZE = DIM*DIM;

    /**
     * The values of this matrix. Named as both a 1d array and a 2d array.
     */
    union {
        real_t m[SIZE];
        real_t _m[DIM][DIM]; // _m[column][row]
    };

    // constructors

    /**
     * Leaves values unitialized.
     */
    Matrix3() {}

    /**
     * Construct a matrix from the given array, in COLUMN MAJOR format
     */
    explicit Matrix3( real_t r[SIZE] );

    /**
     * Construct a matrix from the given values in ROW MAJOR format.
     */
    Matrix3( real_t m00, real_t m10, real_t m20,
             real_t m01, real_t m11, real_t m21,
             real_t m02, real_t m12, real_t m22 );

    static Matrix3 from_columns( const Vector3& x,
                                 const Vector3& y,
                                 const Vector3& z );

    // basic operations
    Matrix3 operator+( const Matrix3& rhs ) const;
    Matrix3& operator+=( const Matrix3& rhs );
    Matrix3 operator-( const Matrix3& rhs ) const;
    Matrix3& operator-=( const Matrix3& rhs );
    Matrix3 operator*( const Matrix3& rhs ) const;
    Vector3 operator*( const Vector3& v ) const;
    Matrix3& operator*=( const Matrix3& rhs );
    Matrix3 operator*( real_t r ) const;
    Matrix3& operator*=( real_t r );
    Matrix3 operator/( real_t r ) const;
    Matrix3& operator/=( real_t r );
    Matrix3 operator-() const;

    // comparaisons
    bool operator==( const Matrix3& rhs ) const;
    bool operator!=( const Matrix3& rhs ) const;

    // accessors

    /**
     * Matrix3(i,j) gives the element at the ith column and jth row.
     */
    real_t operator()( int col, int row ) const {
        assert( col >= 0 && col < DIM && row >= 0 && row < DIM );
        return _m[col][row];
    }

    /**
     * Matrix3(i,j) gives the element at the ith column and jth row.
     */
    real_t& operator()( int col, int row ) {
        assert( col >= 0 && col < DIM && row >= 0 && row < DIM );
        return _m[col][row];
    }

    /**
     * Transform the given vector using this matrix
     */
    Vector3 transform( const Vector3& v ) const {
        return operator*( v );
    }

    /**
     * Combines two transformations into one, with this Matrix being the first
     * to be applied (rhs) and the given matrix the second (lhs). If this matrix
     * is A and the given lhs is B, this function is equivalent to A=B*A.
     */
    Matrix3& concatenate( const Matrix3& lhs ) {
        return *this = lhs * ( *this );
    }

    /**
     * Returns the transpose matrix.
     */
    void transpose( Matrix3& rv ) const;

    /**
     * Returns the inverse matrix. Is undefined if determinant is 0.
     */
    void inverse( Matrix3& rv ) const;
};

inline Matrix3 operator*( real_t r, const Matrix3& m )
{
    return m * r;
}

/**
 * A 4x4 matrix.
 *
 * The matrix is in column-vector compatible notation and is stored
 * in memory by column first. So the memory layout looks like
 *
 * [ m[0] m[4] m[ 8] m[12] ]
 * [ m[1] m[5] m[ 9] m[13] ]
 * [ m[2] m[6] m[10] m[14] ]
 * [ m[3] m[7] m[11] m[15] ]
 *
 * where vectors are treated as column matrices (4x1), i.e. vectors are
 * multiplied on the right of the matrix (M * V)
 *
 * NOTE: The notation mij or m[i][j] is the ith COLUMN and the jth ROW.
 *
 * The matrix is meant for use in linear transformations. Transformations must
 * be combined right-to-left, so if you want to apply A then B the C to a
 * vector V, you would use C*B*A*V.
 */
class Matrix4
{
public:
    /**
     * The identity matrix.
     */
    static const Matrix4 Identity;

    /**
     * The zero matrix.
     */
    static const Matrix4 Zero;

    static const int DIM = 4;
    static const int SIZE = DIM*DIM;

    /**
     * The values of this matrix. Named as both a 1d array and a 2d array.
     */
    union {
        real_t m[SIZE];
        real_t _m[DIM][DIM]; // _m[column][row]
    };

    // constructors

    /**
     * Leaves values unitialized.
     */
    Matrix4() {}

    /**
     * Construct a matrix from the given array, in COLUMN MAJOR format
     */
    explicit Matrix4( real_t r[SIZE] );

    /**
     * Construct a matrix from the given values in ROW MAJOR format (A,B,C,D,E,F,G,H,I).
     */
    Matrix4( real_t m00, real_t m10, real_t m20, real_t m30,
             real_t m01, real_t m11, real_t m21, real_t m31,
             real_t m02, real_t m12, real_t m22, real_t m32,
             real_t m03, real_t m13, real_t m23, real_t m33 );

    // basic operations
    Matrix4 operator+( const Matrix4& rhs ) const;
    Matrix4& operator+=( const Matrix4& rhs );
    Matrix4 operator-( const Matrix4& rhs ) const;
    Matrix4& operator-=( const Matrix4& rhs );
    Matrix4 operator*( const Matrix4& rhs ) const;
    Vector4 operator*( const Vector4& v ) const;
    Matrix4& operator*=( const Matrix4& rhs );
    Matrix4 operator*( real_t r ) const;
    Matrix4& operator*=( real_t r );
    Matrix4 operator/( real_t r ) const;
    Matrix4& operator/=( real_t r );
    Matrix4 operator-() const;

    // comparaisons
    bool operator==( const Matrix4& rhs ) const;
    bool operator!=( const Matrix4& rhs ) const;

    // accessors

    /**
     * Matrix4(i,j) gives the element at the ith column and jth row.
     */
    real_t operator()( int col, int row ) const {
        assert( col >= 0 && col < DIM && row >= 0 && row < DIM );
        return _m[col][row];
    }

    /**
     * Matrix4(i,j) gives the element at the ith column and jth row.
     */
    real_t& operator()( int col, int row ) {
        assert( col >= 0 && col < DIM && row >= 0 && row < DIM );
        return _m[col][row];
    }

    /**
     * Transform the given vector using this matrix
     */
    Vector4 transform( const Vector4& v ) const {
        return operator*( v );
    }

    /**
     * Transforms a Vector3 as a point (fourth component is 1).
     */
    Vector3 transform_point( const Vector3& v ) const {
        return transform( Vector4( v, 1 ) ).projection();
    }

    /**
     * Transforms a Vector3 as a direction (fourth component is 0).
     */
    Vector3 transform_vector( const Vector3& v ) const {
        return transform( Vector4( v, 0 ) ).projection();
    }

    /**
     * Combines two transformations into one, with this Matrix being the first
     * to be applied (rhs) and the given matrix the second (lhs).
     * This is equivalent to *this = lhs * ( *this ).
     */
    Matrix4& concatenate( const Matrix4& lhs ) {
        return *this = lhs * ( *this );
    }

    /**
     * Returns the transpose matrix.
     */
    void transpose( Matrix4& rv ) const;

    /**
     * Returns the inverse matrix. Is undefined if determinant is 0.
     */
    void inverse( Matrix4& rv ) const;
};

inline Matrix4 operator*( real_t r, const Matrix4& m )
{
    return m * r;
}

/**
 * Constructs a matrix that traslates vectors by the given position.
 */
void make_translation_matrix( Matrix4& mat, const Vector3& pos );

/**
 * Constructs a matrix that rotates vectors by the given orientation.
 */
void make_rotation_matrix( Matrix4& mat, const Quaternion& ori );

/**
 * Constructs a matrix that scales vectors by the given scale.
 */
void make_scaling_matrix( Matrix4& mat, const Vector3& scl );

/**
 * Constructs a matrix that scales, rotates, then translates vectors.
 */
void make_transformation_matrix( Matrix4& mat,
                                 const Vector3& pos,
                                 const Quaternion& ori,
                                 const Vector3& scl );

/**
 * Constructs the matrix to transform normal vectors corresponding to the
 * given transformation matrix.
 */
void make_normal_matrix( Matrix3& rv, const Matrix4& tmat );

/**
 * Constructs the matrix to transform normal vectors corresponding to the
 * given transformation matrix.
 */
void make_normal_matrix( Matrix3& rv, const Matrix3& tmat );

} /* _462 */

#endif /* _462_MATH_MATRIX_HPP_ */

