/**
 * @file quaternion.hpp
 * @brief A quaternion class.
 *
 * @author Eric Butler (edbutler)
 * @author Zeyang Li (zeyangl)
 */

#ifndef _462_MATH_QUATERNION_HPP_
#define _462_MATH_QUATERNION_HPP_

#include "math/math.hpp"
#include "math/vector.hpp"

namespace _462 {

class Matrix3;
class Matrix4;

/*
   This code is loosely based on quaternion code from Ogre3d (www.ogre3d.org).
 */

class Quaternion
{
public:

    /**
     * The quaternion filled with zero.
     */
    static const Quaternion Zero;

    /**
     * The quaternion representing the identity rotation.
     */
    static const Quaternion Identity;

    real_t w, x, y, z;

    /**
     * Default constructor. Leaves values uninitialized.
     */
    Quaternion() {}

    /**
     * Construct a quaternion with the given values.
     */
    Quaternion( real_t w, real_t x, real_t y, real_t z )
        : w( w ), x( x ), y( y ), z( z ) { }

    /**
     * Constructs a quaternion representing a rotation about the given axis
     * by the given angle.
     */
    Quaternion( const Vector3& axis, real_t radians );

    /**
     *  Constructs a quaternion from a rotation matrix.
     */
    explicit Quaternion( const Matrix4& mat );

    Quaternion operator*( const Quaternion& rhs ) const;

    /**
     * Rotate a vector by this quaternion.
     */
    Vector3 operator*( const Vector3& rhs ) const;

    bool operator==( const Quaternion& rhs ) const {
        return rhs.x == x && rhs.y == y &&
               rhs.z == z && rhs.w == w;
    }

    bool operator!=( const Quaternion& rhs ) const {
        return !operator==( rhs );
    }

    real_t norm() const {
        return x * x + y * y + z * z + w * w;
    }

    real_t magnitude() const {
        return sqrt( norm() );
    }

    Quaternion& normalize();

    /**
     * Returns the inverse of this quaternion.
     */
    Quaternion inverse() const;

    /**
     * Convert this quaternion into an angle and axis.
     * Returns the rotation in radians about an axis.
     */
    void to_angle_axis( Vector3& axis, real_t* angle ) const;

    /**
     * Converts this quaternion to a 3x3 matrix.
     */
    void to_matrix( Matrix3& mat ) const;

    /**
     * Converts this quaternion to a 4x4 matrix.
     */
    void to_matrix( Matrix4& mat ) const;

    /**
     * Returns the X,Y,Z axes rotated by this quaternion.
     */
    void to_axes( Vector3 axes[3] ) const;
};

std::ostream& operator <<( std::ostream& o, const Quaternion& q );

inline Vector3 operator*( const Vector3& v, const Quaternion& q )
{
    return q * v;
}

inline Vector3& operator*=( Vector3& v, const Quaternion& q )
{
    return v = q * v;
}

} /* _462 */

#endif /* _462_MATH_QUATERNION_HPP_ */

