/**
 * @file spline.hpp
 * @brief Spline classes
 *
 * @author Kristin Siu
 */

#ifndef _462_MATH_SPLINE_HPP_
#define _462_MATH_SPLINE_HPP_

#include "math/math.hpp"
#include <algorithm>
#include <cassert>
#include <cmath>
#include <fstream>

namespace _462{

/**
 * A spline
 */
template < typename V >
class Spline
{
public:

    /**
     * The coefficients of this spline
     */
    V a, b, c, d;

    /**
     * Default constructor.
     * Leaves coefficients unintialized
     */
    Spline() { }

    /**
     * @brief Sets the spline coefficients
     * @remark This is for Catmull-Rom splines
     * @param p0, p1, p2, p3 The control points
     * @param tau The tension value.  Affects how sharply curve bends.
     */
    void set_spline_coefficients( V p0, V p1, V p2, V p3, real_t tau )
    {
        a = ( ( -1.0 * tau ) * p0) + ( ( 2.0 - tau ) * p1 ) + ( (        tau - 2.0 ) * p2 ) + (        tau * p3 );
        b = ( (  2.0 * tau ) * p0) + ( ( tau - 3.0 ) * p1 ) + ( ( -2.0 * tau + 3.0 ) * p2 ) + ( -1.0 * tau * p3 );
        c = ( ( -1.0 * tau ) * p0) + ( tau * p2 );
        d = p1;
    }

    void set_spline_coefficients( const V* arr, real_t tau )
    {
        set_spline_coefficients( arr[0], arr[1], arr[2], arr[3], tau );
    }

    /**
     * @brief Evaluates the spline at a given time
     * @param t Time
     */
    V eval( real_t t ) const
    {
        //return (a * (t * t * t)) + (b * (t * t)) + (c * t) + d;
        return t * ( t * ( a * t + b ) + c ) + d; // optimization of the above
    }

    /**
     * @brief Calculates the derivative at the given point
     * @param t Time
     */
    V deriv( real_t t ) const
    {
        //return (3 * a * (t * t)) + (2 * b * t) + c;
        return t * ( 3 * a * t + 2 * b ) + c; // optimization of the above
    }
};

} /* _462 */

#endif /* _462_MATH_SPLINE_HPP_ */

