/**
 * @file vector.hpp
 * @brief Vector classes.
 *
 * @author Eric Butler (edbutler)
 */

#ifndef _462_MATH_VECTOR_HPP_
#define _462_MATH_VECTOR_HPP_

#include "math/math.hpp"
#include <algorithm>
#include <cassert>
#include <cmath>
#include <fstream>

namespace _462 {

/**
 * A 2d vector.
 */
class Vector2
{
public:

    /**
     * The number of dimensions.
     */
    static const size_t DIM = 2;

    /**
     * The zero vector.
     */
    static const Vector2 Zero;

    /**
     * The vector (1,1).
     */
    static const Vector2 Ones;

    /**
     * The vector (1,0).
     */
    static const Vector2 UnitX;

    /**
     * The vector (0,1).
     */
    static const Vector2 UnitY;

    /**
     * Components of this vector.
     */
    real_t x, y;

    /**
     * Default constructor. Leaves values unitialized.
     */
    Vector2() {}

    /**
     * Create a vector with the given values.
     */
    Vector2( real_t x, real_t y )
		: x( x ), y( y ) {}

    /**
     * Create a vector from the elements of the given array.
     */
    explicit Vector2( const float p[DIM] )
		: x( p[0] ), y( p[1] ) {}

    /**
     * Create a vector from the elements of the given array.
     */
    explicit Vector2( const double p[DIM] )
        : x( p[0] ), y( p[1] ) {}

    // also uses default copy and assignment

    Vector2 operator+( const Vector2& rhs ) const {
        return Vector2( x + rhs.x, y + rhs.y );
    }

    Vector2& operator+=( const Vector2& rhs ) {
        x += rhs.x;
        y += rhs.y;
        return *this;
    }

    Vector2 operator-( const Vector2& rhs ) const {
        return Vector2( x - rhs.x, y - rhs.y );
    }

    Vector2& operator-=( const Vector2& rhs ) {
        x -= rhs.x;
        y -= rhs.y;
        return *this;
    }

    // not really mathematically sound, but makes sense for colors
    Vector2 operator*( const Vector2& rhs ) const {
        return Vector2( x * rhs.x, y * rhs.y );
    }

    // not really mathematically sound, but makes sense for colors
    Vector2& operator*=( const Vector2& rhs ) {
        x *= rhs.x;
        y *= rhs.y;
        return *this;
    }

    Vector2 operator*( real_t s ) const {
        return Vector2( x * s, y * s );
    }

    Vector2& operator*=( real_t s ) {
        x *= s;
        y *= s;
        return *this;
    }

    Vector2 operator/( real_t s ) const {
        real_t inv = 1.0 / s;
        return Vector2( x * inv, y * inv );
    }

    Vector2& operator/=( real_t s ) {
        real_t inv = 1.0 / s;
        x *= inv;
        y *= inv;
        return *this;
    }

    Vector2 operator-() const {
        return Vector2( -x, -y );
    }

    /**
     * @remark No bounds checking.
     */
    real_t operator[]( size_t i ) const {
		// assumes all members are in a contiguous block
        assert( i >= 0 && i < DIM );
        return ( &x )[i];
    }

    /**
     * @remark No bounds checking.
     */
    real_t& operator[]( size_t i ) {
		// assumes all members are in a contiguous block
        assert( i >= 0 && i < DIM );
        return ( &x )[i];
    }

    bool operator==( const Vector2& rhs ) const {
        return x == rhs.x && y == rhs.y;
    }

    bool operator!=( const Vector2& rhs ) const {
        return !operator==( rhs );
    }

    /**
     * Returns the dot product of two vectors
     */
    real_t dot( const Vector2& rhs ) const {
        return x * rhs.x + y * rhs.y;
    }

    /**
     * Returns the magnitude of a vector
     */
    real_t magnitude() const {
        return sqrt( x * x + y * y );
    }

    /**
     * Efficiency function: does not require square root operation.
     */
    real_t squared_magnitude() const {
        return x * x + y * y;
    }

    /**
     * Calculate the positive distance between two vectors.
     */
    real_t distance( const Vector2& rhs ) const {
        return ( *this - rhs ).magnitude();
    }

    /**
     * Efficiency function: does not require square root operation.
     */
    real_t squared_distance( const Vector2& rhs ) const {
        return ( *this - rhs ).squared_magnitude();
    }

    /**
     * Returns the unit vector pointing in the same direction as this vector.
     */
    Vector2 unit() const {
        return *this / magnitude();
    }

    /**
     * Noramlizes this vector; that is, sets its magnitude to 1.
     */
    Vector2& normalize() {
        return *this /= magnitude();
    }

    /**
     * Returns a vector whose elements are the absolute values of all the
     * elements of this vector.
     */
    Vector2 abs() const {
        return Vector2( fabs( x ), fabs( y ) );
    }

    /**
     * Returns a vector which is the point exactly between this and the given
     * vector.
     */
    Vector2 midpoint( const Vector2& rhs ) const {
        return ( *this + rhs ) * .5;
    }

    /**
     * Clamps the lower bound of this vector; that is, sets this vector's values
     * to the max of the current values and the given vector's values.
     * @param rhs The vector representing the desired lower bound.
     */
    Vector2& clamp_min( const Vector2& rhs ) {
        x = std::max( x, rhs.x );
        y = std::max( y, rhs.y );
        return *this;
    }

    /**
     * Clamps the upper bound of this vector; that is, sets this vector's values
     * to the min of the current values and the given vector's values.
     * @param rhs The vector representing the desired upper bound.
     */
    Vector2& clamp_max( const Vector2& rhs ) {
        x = std::min( x, rhs.x );
        y = std::min( y, rhs.y );
        return *this;
    }

    /**
     * Returns a vector whose values are the maximum of this vector's values and
     * the given vector's values.
     */
    Vector2 maximum( const Vector2& rhs ) const {
		Vector2 rv( *this );
		rv.clamp_min( rhs );
		return rv;
    }

    /**
     * Returns a vector whose values are the minimum of this vector's values and
     * the given vector's values.
     */
    Vector2 minimum( const Vector2& rhs ) const {
		Vector2 rv( *this );
		rv.clamp_max( rhs );
		return rv;
    }

    void to_array( float arr[DIM] ) const {
        arr[0] = float( x );
        arr[1] = float( y );
    }

    void to_array( double arr[DIM] ) const {
        arr[0] = double( x );
        arr[1] = double( y );
    }
};

inline Vector2 operator*( real_t s, const Vector2& rhs )
{
    return rhs * s;
}

/**
 * Outputs a vector text formatted as "(x,y,z)".
 */
std::ostream& operator<<( std::ostream& os, const Vector2& rhs );

/**
 * A 3d vector.
 */
class Vector3
{
public:

    /**
     * The number of dimensions.
     */
    static const size_t DIM = 3;

    /**
     * The zero vector.
     */
    static const Vector3 Zero;

    /**
     * The vector (1,1,1).
     */
    static const Vector3 Ones;

    /**
     * The vector (1,0,0)
     */
    static const Vector3 UnitX;

    /**
     * The vector (0,1,0)
     */
    static const Vector3 UnitY;

    /**
     * The vector (0,0,1)
     */
    static const Vector3 UnitZ;

    /**
     * Components of this vector.
     */
    real_t x, y, z;

    /**
     * Default constructor. Leaves values unitialized.
     */
    Vector3() {}

    /**
     * Create a vector with the given values.
     */
    Vector3( real_t x, real_t y, real_t z )
		: x( x ), y( y ), z( z ) {}

    /**
     * Create a vector from a 2d vector.
     */
    Vector3( const Vector2& v, real_t z )
        : x( v.x ), y( v.y ), z( z ) { }

    /**
     * Create a vector from the elements of the given array.
     */
    explicit Vector3( const float p[DIM] )
		: x( p[0] ), y( p[1] ), z( p[2] ) {}

    /**
     * Create a vector from the elements of the given array.
     */
    explicit Vector3( const double p[DIM] )
		: x( p[0] ), y( p[1] ), z( p[2] ) {}

    // also uses default copy and assignment

    Vector3 operator+( const Vector3& rhs ) const {
        return Vector3( x + rhs.x, y + rhs.y, z + rhs.z );
    }

    Vector3& operator+=( const Vector3& rhs ) {
        x += rhs.x;
        y += rhs.y;
        z += rhs.z;
        return *this;
    }

    Vector3 operator-( const Vector3& rhs ) const {
        return Vector3( x - rhs.x, y - rhs.y, z - rhs.z );
    }

    Vector3& operator-=( const Vector3& rhs ) {
        x -= rhs.x;
        y -= rhs.y;
        z -= rhs.z;
        return *this;
    }

    // not really mathematically sound, but makes sense for colors
    Vector3 operator*( const Vector3& rhs ) const {
        return Vector3( x * rhs.x, y * rhs.y, z * rhs.z );
    }

    // not really mathematically sound, but makes sense for colors
    Vector3& operator*=( const Vector3& rhs ) {
        x *= rhs.x;
        y *= rhs.y;
        z *= rhs.z;
        return *this;
    }

    Vector3 operator*( real_t s ) const {
        return Vector3( x * s, y * s, z * s );
    }

    Vector3& operator*=( real_t s ) {
        x *= s;
        y *= s;
        z *= s;
        return *this;
    }

    Vector3 operator/( real_t s ) const {
        real_t inv = 1.0 / s;
        return Vector3( x * inv, y * inv, z * inv );
    }

    Vector3& operator/=( real_t s ) {
        real_t inv = 1.0 / s;
        x *= inv;
        y *= inv;
        z *= inv;
        return *this;
    }

    Vector3 operator-() const {
        return Vector3( -x, -y, -z );
    }

    /**
     * @remark No bounds checking.
     */
    real_t operator[]( size_t i ) const {
		// assumes all members are in a contiguous block
        assert( i >= 0 && i < DIM );
        return ( &x )[i];
    }

    /**
     * @remark No bounds checking.
     */
    real_t& operator[]( size_t i ) {
		// assumes all members are in a contiguous block
        assert( i >= 0 && i < DIM );
        return ( &x )[i];
    }

    bool operator==( const Vector3& rhs ) const {
        return x == rhs.x && y == rhs.y && z == rhs.z;
    }

    bool operator!=( const Vector3& rhs ) const {
        return !operator==( rhs );
    }

    /**
     * Returns the dot product of two vectors
     */
    real_t dot( const Vector3& rhs ) const {
        return x * rhs.x + y * rhs.y + z * rhs.z;
    }

    /**
     * Returns the cross product of two vectors
     */
    Vector3 cross( const Vector3& rhs ) const {
        return Vector3(
			y * rhs.z - z * rhs.y,
			z * rhs.x - x * rhs.z,
			x * rhs.y - y * rhs.x
		);
    }

    /**
     * Returns the magnitude of a vector
     */
    real_t magnitude() const {
        return sqrt( x * x + y * y + z * z );
    }

    /**
     * Efficiency function: does not require square root operation.
     */
    real_t squared_magnitude() const {
        return x * x + y * y + z * z;
    }

    /**
     * Calculate the positive distance between two vectors.
     */
    real_t distance( const Vector3& rhs ) const {
        return ( *this - rhs ).magnitude();
    }

    /**
     * Efficiency function: does not require square root operation.
     */
    real_t squared_distance( const Vector3& rhs ) const {
        return ( *this - rhs ).squared_magnitude();
    }

    /**
     * Returns the unit vector pointing in the same direction as this vector.
     */
    Vector3 unit() const {
        return *this / magnitude();
    }

    /**
     * Noramlizes this vector; that is, sets its magnitude to 1.
     */
    Vector3& normalize() {
        return *this /= magnitude();
    }

    /**
     * Returns a vector whose elements are the absolute values of all the
     * elements of this vector.
     */
    Vector3 abs() const {
        return Vector3( fabs( x ), fabs( y ), fabs( z ) );
    }

    /**
     * Returns a vector which is the point exactly between this and the given
     * vector.
     */
    Vector3 midpoint( const Vector3& rhs ) const {
        return ( *this + rhs ) * .5;
    }

    /**
     * Clamps the lower bound of this vector; that is, sets this vector's values
     * to the max of the current values and the given vector's values.
     * @param rhs The vector representing the desired lower bound.
     */
    Vector3& clamp_min( const Vector3& rhs ) {
        x = std::max( x, rhs.x );
        y = std::max( y, rhs.y );
        z = std::max( z, rhs.z );
        return *this;
    }

    /**
     * Clamps the upper bound of this vector; that is, sets this vector's values
     * to the min of the current values and the given vector's values.
     * @param rhs The vector representing the desired upper bound.
     */
    Vector3& clamp_max( const Vector3& rhs ) {
        x = std::min( x, rhs.x );
        y = std::min( y, rhs.y );
        z = std::min( z, rhs.z );
        return *this;
    }

    /**
     * Returns a vector whose values are the maximum of this vector's values and
     * the given vector's values.
     */
    Vector3 maximum( const Vector3& rhs ) const {
		Vector3 rv( *this );
		rv.clamp_min( rhs );
		return rv;
    }

    /**
     * Returns a vector whose values are the minimum of this vector's values and
     * the given vector's values.
     */
    Vector3 minimum( const Vector3& rhs ) const {
		Vector3 rv( *this );
		rv.clamp_max( rhs );
		return rv;
    }

    void to_array( float arr[DIM] ) const {
        arr[0] = float( x );
        arr[1] = float( y );
        arr[2] = float( z );
    }

    void to_array( double arr[DIM] ) const {
        arr[0] = double( x );
        arr[1] = double( y );
        arr[2] = double( z );
    }
};

inline Vector3 operator*( real_t s, const Vector3& rhs )
{
    return rhs * s;
}

/**
 * Outputs a vector text formatted as "(x,y,z)".
 */
std::ostream& operator<<( std::ostream& os, const Vector3& rhs );

/**
 * A 4d vector.
 */
class Vector4
{
public:

    /**
     * The number of dimensions.
     */
    static const size_t DIM = 4;

    /**
     * The zero vector.
     */
    static const Vector4 Zero;

    /**
     * The vector (1,1,1,1).
     */
    static const Vector4 Ones;

    /**
     * The vector (1,0,0,0)
     */
    static const Vector4 UnitX;

    /**
     * The vector (0,1,0,0)
     */
    static const Vector4 UnitY;

    /**
     * The vector (0,0,1,0)
     */
    static const Vector4 UnitZ;

    /**
     * The vector (0,0,0,1)
     */
    static const Vector4 UnitW;

    /**
     * Components of this vector.
     */
    real_t x, y, z, w;

    /**
     * Default constructor. Leaves values unitialized.
     */
    Vector4() {}

    /**
     * Create a vector with the given values.
     */
    Vector4( real_t x, real_t y, real_t z, real_t w )
		: x( x ), y( y ), z( z ), w( w ) {}

    /**
     * Create the vector (v.x, v.y, v.z, w).
     */
    Vector4( const Vector3& v, real_t w )
		: x( v.x ), y( v.y ), z( v.z ), w( w ) {}

    /**
     * Create a vector from the elements of the given array.
     */
    explicit Vector4( const float p[DIM] )
		: x( p[0] ), y( p[1] ), z( p[2] ), w( p[3] ) {}

    /**
     * Create a vector from the elements of the given array.
     */
    explicit Vector4( const double p[DIM] )
		: x( p[0] ), y( p[1] ), z( p[2] ), w( p[3] ) {}

    // also uses default copy and assignment

    Vector4 operator+( const Vector4& rhs ) const {
        return Vector4( x + rhs.x, y + rhs.y, z + rhs.z, w + rhs.w );
    }

    Vector4& operator+=( const Vector4& rhs ) {
        x += rhs.x;
        y += rhs.y;
        z += rhs.z;
        w += rhs.w;
        return *this;
    }

    Vector4 operator-( const Vector4& rhs ) const {
        return Vector4( x - rhs.x, y - rhs.y, z - rhs.z, w - rhs.w );
    }

    Vector4& operator-=( const Vector4& rhs ) {
        x -= rhs.x;
        y -= rhs.y;
        z -= rhs.z;
        w -= rhs.w;
        return *this;
    }

    // not really mathematically sound, but makes sense for colors
    Vector4 operator*( const Vector4& rhs ) const {
        return Vector4( x * rhs.x, y * rhs.y, z * rhs.z, w * rhs.w );
    }

    // not really mathematically sound, but makes sense for colors
    Vector4& operator*=( const Vector4& rhs ) {
        x *= rhs.x;
        y *= rhs.y;
        z *= rhs.z;
        w *= rhs.w;
        return *this;
    }

    Vector4 operator*( real_t s ) const {
        return Vector4( x * s, y * s, z * s, w * s );
    }

    Vector4& operator*=( real_t s ) {
        x *= s;
        y *= s;
        z *= s;
        w *= s;
        return *this;
    }

    Vector4 operator/( real_t s ) const {
        real_t inv = 1.0 / s;
        return Vector4( x * inv, y * inv, z * inv, w * inv );
    }

    Vector4& operator/=( real_t s ) {
        real_t inv = 1.0 / s;
        x *= inv;
        y *= inv;
        z *= inv;
        w *= inv;
        return *this;
    }

    Vector4 operator-() const {
        return Vector4( -x, -y, -z, -w );
    }

    /**
     * @remark No bounds checking.
     */
    real_t operator[]( size_t i ) const {
		// assumes all members are in a contiguous block
        assert( i >= 0 && i < DIM );
        return ( &x )[i];
    }

    /**
     * @remark No bounds checking.
     */
    real_t& operator[]( size_t i ) {
		// assumes all members are in a contiguous block
        assert( i >= 0 && i < DIM );
        return ( &x )[i];
    }

    bool operator==( const Vector4& rhs ) const {
        return x == rhs.x && y == rhs.y && z == rhs.z && w == rhs.w;
    }

    bool operator!=( const Vector4& rhs ) const {
        return !operator==( rhs );
    }

    /**
     * Returns the 3d vector corresponding to this 4d vector.
     * @remark If w==0, returns (x,y,z).
     */
    Vector3 projection() const {
        real_t tmpwinv = w == 0.0 ? 1.0 : 1.0 / w;
        return Vector3( x * tmpwinv, y * tmpwinv, z * tmpwinv );
    }

    /**
     * Returns the first three components, ignoring the fourth
     */
    Vector3 xyz() const {
        return Vector3( x, y, z );
    }

    /**
     * Returns the dot product of two vectors
     */
    real_t dot( const Vector4& rhs ) const {
        return x * rhs.x + y * rhs.y + z * rhs.z + w * rhs.w;
    }

    /**
     * Returns the magnitude of a vector
     */
    real_t magnitude() const {
        return sqrt( x*x + y*y + z*z + w*w );
    }

    /**
     * Efficiency function: does not require square root operation.
     */
    real_t squared_magnitude() const {
        return x*x + y*y + z*z + w*w;
    }

    /**
     * Calculate the positive distance between two vectors.
     */
    real_t distance( const Vector4& rhs ) const {
        return ( *this - rhs ).magnitude();
    }

    /**
     * Efficiency function: does not require square root operation.
     */
    real_t squared_distance( const Vector4& rhs ) const {
        return ( *this - rhs ).squared_magnitude();
    }

    /**
     * Returns the unit vector pointing in the same direction as this vector.
     */
    Vector4 unit() const {
        return *this / magnitude();
    }

    /**
     * Noramlizes this vector; that is, sets its magnitude to 1.
     */
    Vector4& normalize() {
        return *this /= magnitude();
    }

    /**
     * Returns a vector whose elements are the absolute values of all the
     * elements of this vector.
     */
    Vector4 abs() const {
        return Vector4( fabs( x ), fabs( y ), fabs( z ), fabs( w ) );
    }

    /**
     * Returns a vector which is the point exactly between this and the given
     * vector.
     */
    Vector4 midpoint( const Vector4& rhs ) const {
        return ( *this + rhs ) * .5;
    }

    /**
     * Clamps the lower bound of this vector; that is, sets this vector's values
     * to the max of the current values and the given vector's values.
     * @param rhs The vector representing the desired lower bound.
     */
    Vector4& clamp_min( const Vector4& rhs ) {
        x = std::max( x, rhs.x );
        y = std::max( y, rhs.y );
        z = std::max( z, rhs.z );
        w = std::max( w, rhs.w );
        return *this;
    }

    /**
     * Clamps the upper bound of this vector; that is, sets this vector's values
     * to the min of the current values and the given vector's values.
     * @param rhs The vector representing the desired upper bound.
     */
    Vector4& clamp_max( const Vector4& rhs ) {
        x = std::min( x, rhs.x );
        y = std::min( y, rhs.y );
        z = std::min( z, rhs.z );
        w = std::min( w, rhs.w );
        return *this;
    }

    /**
     * Returns a vector whose values are the maximum of this vector's values and
     * the given vector's values.
     */
    Vector4 maximum( const Vector4& rhs ) const {
		Vector4 rv( *this );
		rv.clamp_min( rhs );
		return rv;
    }

    /**
     * Returns a vector whose values are the minimum of this vector's values and
     * the given vector's values.
     */
    Vector4 minimum( const Vector4& rhs ) const {
		Vector4 rv( *this );
		rv.clamp_max( rhs );
		return rv;
    }

    void to_array( float arr[DIM] ) const {
        arr[0] = float( x );
        arr[1] = float( y );
        arr[2] = float( z );
        arr[3] = float( w );
    }

    void to_array( double arr[DIM] ) const {
        arr[0] = double( x );
        arr[1] = double( y );
        arr[2] = double( z );
        arr[3] = double( w );
    }
};

inline Vector4 operator*( real_t s, const Vector4& rhs )
{
    return rhs * s;
}

/**
 * Outputs a vector text formatted as "(x,y,z)".
 */
std::ostream& operator<<( std::ostream& os, const Vector4& rhs );

} /* _462 */

#endif /* _462_MATH_VECTOR_HPP_ */

